* {
    outline: none;
}

Window {
    min-width: 580px;
    min-height: 300px;
    height: 350px;
    background: #eee;
}

Window #header {
    background: white;
    border-bottom: 1px solid #ccc;
}

Window #pages QTableView {
    border: 0;
    margin-bottom: 5px;
}

Window #apps {
    min-width: 300px;
}

Window #loading QLabel {
    font-size: 16px;
}

Window App #tools {
    max-height: 100px;
    min-height: 100px;
}

Window App #label {
    font-size: 13px;
}

Window App QAbstractItemView {
    max-height: 100px;
}

QHeaderView {
    font-weight: bold;
}

QHeaderView {
    qproperty-defaultAlignment: AlignLeft;
    font-weight: bold;
}

QMenu QLabel {
    padding-top: 5px;
    padding-bottom: 5px;
    font-weight: bold;
}

QHeaderView::section {
    padding: 0px;
    padding-left: 5px;
    padding-right: 5px;
    padding-top: 5px;
    background: #f5f5f5;
    border: 0;
    border-right: 1px solid #fff;
    border-left: 1px solid #e5e5e5
}

QHeaderView::section:first {
    border-left: 0;
}

QHeaderView::section:last {
        border-right: 0;
}

/* Docks */
Context #generateGraph {
   max-width: 60px;
    min-width: 60px;
    max-height: 20px;
}

/* Risky, dependent on font and plaform */
Context #overlay {
    max-width: 200px;
    min-width: 200px;
    max-height: 150px;
    min-height: 150px;
}

AbstractDockWidget #dockHelp {
    background: transparent;
    border: none;
    padding-left: 5px;
    padding-bottom: 5px;
    padding-right: 5px;
}

Window #header QLabel {
    padding: 0;
}

Window #header QComboBox {
    padding: 0;
}

Window #header #profileBtn {
    background: transparent;
    border: 1px solid #aaa;
    padding: 0;
}

Window #header #profileReset {

}

Window QStatusBar #profileReset {
    max-width: 25px;
    padding: 0;
    margin: 0;
    max-width: 16px;
    max-height: 16px;
    border: 5px solid yellow;
    background: green;
}


Window QStatusBar #profileReset:hover {
    background: blue;
}


Window QStatusBar #profileReset:pressed {
    background: yellow;
}

Window #body {
    background: white;
}

FullCommand QLineEdit {
  color: rgba(0, 0, 0, 0.2);
}

FullCommand QComboBox {
    background: white;
    color: #222;
}

FullCommand QComboBox {
    border: none;
    padding: 0;
    padding-left: 3px;
}

FullCommand QComboBox::drop-down {
    border: none;
    width: 0;
}

FullCommand QComboBox::down-arrow {
    image: url(%(res)s/logo_64.png);
}

Window #body QLineEdit {
    border: 1px solid #ddd;
    padding: 3px;
    margin: 3px;
}

LineEditWithCompleter {
    background: none;
    border: none;

    /* Leave room for hover */
    padding-bottom: 2px;
}

/* Underline on hover */
LineEditWithCompleter:hover {
    border-bottom: 2px solid steelblue;
    padding: 0px;
}

Window #header QPushButton[type=toggle] {
}

Window #logo {
    width: 32px;
    max-width: 32px;
    height: 32px;
    max-height: 32px;
    padding: 0;
}

/* Give it a slight nudge */
Window #logo:hover {
    border: -1px solid transparent;
}

Console {
    min-width: 200px;
    min-height: 100px;
}

Options QTableView {
    background: transparent;
    border: none;
}

Options #label {
    font-size: 15px;
}

QToolBar {
    border: none;   
}

QToolButton {
    min-height: 35px;
}

Window #header #profileBtn {
    padding-left: -2px;
    padding-right: -3px;
    border: 2px solid #333;
}

Window #header #profileBtn:hover {
    border: 2px solid steelblue;
}

Window #header #profileBtn:pressed {
    border: 1px solid white;
}

QTableView {
    /*background: transparent;*/
    /*border: none;*/
}

#launchBtn {
    height: 30px;
    margin: 0;
    padding: 0;
}

#launchOptionsBtn {
    width: 30px;
    max-height: 30px;
    min-height: 30px;
    min-width: 30px;
}

#busyIndicator {
    min-height: 5px;
    max-height: 5px;
    background: #ef6363;
}

QTableView {
    selection-background-color: #ddd;
    selection-color: #111;
}

QTableView::item:selected {
    border: none;
}

QTableView::item:selected:!active {
    background: transparent;
}

QTableView::item:selected:active, QTableView::item:selected:!focus {
    background: #add9f2;
}

QFrame {
    border: none;
}